@@ -1,30 +1,29 @@
+import inspect
 import pydoc
 import signal
 from typing import Any, Optional, Union
 
 import blessed
-from blessed import Terminal
 import rich
+from blessed import Terminal
 from rich.console import Console
 from rich.layout import Layout
 from rich.panel import Panel
 from rich.syntax import Syntax
+from rich.text import Text
 
-from .utils import is_selectable
 from .cached_object import CachedObject
 from .explorer_layout import ExplorerLayout, ExplorerState
+from .filter_layout import FilterLayout
 from .help_layout import HelpLayout, HelpState, random_error_quote
 from .overview_layout import OverviewLayout, OverviewState, PreviewState
 from .stack_layout import StackFrame, StackLayout
+from .utils import is_selectable
 
-version = "1.2.7"
+version = "1.3.0"
 
-# TODO fix explore(namedtuple)
-# TODO methods filter
-# or just a type filter?
-# TODO empty overview layouts for when there are 0 public attributes
 # TODO search filter
-# TODO use inspect.ismethod inspect.ismodule, inspect.isfunction isclass
+
 
 console = Console()
 
@@ -32,10 +31,10 @@ console = Console()
 class Explorer:
     """ Explorer class used to interactively explore Python Objects """
 
-    def __init__(self, obj: Any):
-        cached_obj = CachedObject(obj, dotpath=repr(obj))
+    def __init__(self, obj: Any, name_of_obj: str):
+        cached_obj = CachedObject(obj, attr_name=name_of_obj)
         # Figure out all the attributes of the current obj's attributes
-        cached_obj.cache_attributes()
+        cached_obj.cache()
 
         # self.head_obj = cached_obj
         self.cached_obj: CachedObject = cached_obj
@@ -44,6 +43,7 @@ class Explorer:
         self.help_layout = HelpLayout(version, visible=False, ratio=3)
         self.explorer_layout = ExplorerLayout(cached_obj=cached_obj)
         self.overview_layout = OverviewLayout(ratio=3)
+        self.filter_layout = FilterLayout(visible=False)
 
         self.stack.append(
             StackFrame(
@@ -97,7 +97,7 @@ class Explorer:
 
         if key == "b":
             breakpoint()
-            pass
+            return
 
         # Help page ###########################################################
 
@@ -123,7 +123,7 @@ class Explorer:
                     self.help_layout.state = HelpState.keybindings
                 return
 
-            elif key in ("j", "k") or key.code in (
+            elif key in ("j", "k", "o", "n") or key.code in (
                 self.term.KEY_UP,
                 self.term.KEY_DOWN,
             ):
@@ -139,39 +139,40 @@ class Explorer:
 
         # Navigation ##########################################################
 
-        # if the stack view is open, only accept inputs to move around/close thes tack view
-        if self.stack.visible and (
-            key not in ("o", "j", "k", "\n")
-            and key.code
-            not in (self.term.KEY_UP, self.term.KEY_DOWN, self.term.KEY_RIGHT)
-        ):
-            return
-
-        elif key == "k" or key.code == self.term.KEY_UP:
+        if key == "k" or key.code == self.term.KEY_UP:
             if self.stack.visible:
                 self.stack.move_up()
+            elif self.filter_layout.visible:
+                self.filter_layout.move_up()
             else:
                 self.explorer_layout.move_up()
 
         elif key == "j" or key.code == self.term.KEY_DOWN:
             if self.stack.visible:
                 self.stack.move_down(self.panel_height)
+            elif self.filter_layout.visible:
+                self.filter_layout.move_down()
             else:
                 self.explorer_layout.move_down(self.panel_height, self.cached_obj)
 
-        # Enter
-        elif key in ("\n", "l") or key.code == self.term.KEY_RIGHT:
+        elif key in ("\n", " ") and self.filter_layout.visible:
+            self.filter_layout.toggle(cached_obj=self.cached_obj)
+
+        elif key in ("\n", "l", " ") or key.code == self.term.KEY_RIGHT:
 
             if self.stack.visible:
+                # If you are choosing the same frame as the current obj, then don't do anything
+                if self.stack[self.stack.index].cached_obj == self.cached_obj:
+                    return
                 new_cached_obj = self.stack.select()
             else:
-                new_cached_obj = self.cached_obj.selected_cached_obj
+                new_cached_obj = self.explorer_layout.selected_object
                 if not is_selectable(new_cached_obj.obj):
                     return
 
             self.explorer_layout = ExplorerLayout(cached_obj=new_cached_obj)
             self.cached_obj = new_cached_obj
-            self.cached_obj.cache_attributes()
+            self.cached_obj.cache()
             self.stack.append(
                 StackFrame(
                     cached_obj=self.cached_obj,
@@ -195,11 +196,26 @@ class Explorer:
 
         # View ################################################################
 
-        if key == "o" and self.stack.visible:
-            self.stack.visible = False
+        if key == "o":
+            if self.stack.visible:
+                self.stack.visible = False
+            elif self.filter_layout.visible:
+                self.filter_layout.visible = False
+                self.stack.set_visible()
+            else:
+                self.stack.set_visible()
 
-        elif key == "o" and not self.stack.visible:
-            self.stack.set_visible()
+        elif key == "n":
+            if self.filter_layout.visible:
+                self.filter_layout.visible = False
+            elif self.stack.visible:
+                self.stack.visible = False
+                self.filter_layout.visible = True
+            else:
+                self.filter_layout.visible = True
+
+        elif key == "c" and self.filter_layout.visible:
+            self.filter_layout.clear_filters(self.cached_obj)
 
         # Switch between public and private attributes
         elif key in ("[", "]"):
@@ -210,7 +226,7 @@ class Explorer:
                 self.explorer_layout.state = ExplorerState.public
 
         elif key in ("{", "}"):
-            if not callable(self.cached_obj.selected_cached_obj.obj):
+            if not callable(self.explorer_layout.selected_object.obj):
                 return
 
             if self.overview_layout.preview_state == PreviewState.repr:
@@ -236,16 +252,16 @@ class Explorer:
 
         # Fullscreen
         elif key == "f":
-            printable: Union[str, Syntax]
+            printable: Union[str, Syntax, Text]
 
             if self.overview_layout.state == OverviewState.docstring:
-                printable = self.cached_obj.selected_cached_obj.docstring
+                printable = self.explorer_layout.selected_object.docstring
 
             elif self.overview_layout.preview_state == PreviewState.repr:
-                printable = self.cached_obj.selected_cached_obj.obj
+                printable = self.explorer_layout.selected_object.obj
 
             elif self.overview_layout.preview_state == PreviewState.source:
-                printable = self.cached_obj.selected_cached_obj.get_source(
+                printable = self.explorer_layout.selected_object.get_source(
                     fullscreen=True
                 )
 
@@ -256,12 +272,12 @@ class Explorer:
             pydoc.pager(str_out)
 
         elif key == "H":
-            help(self.cached_obj.selected_cached_obj.obj)
+            help(self.explorer_layout.selected_object)
 
         elif key == "i":
             with console.capture() as capture:
                 rich.inspect(
-                    self.cached_obj.selected_cached_obj.obj,
+                    self.explorer_layout.selected_object.obj,
                     console=console,
                     methods=True,
                 )
@@ -271,7 +287,7 @@ class Explorer:
         elif key == "I":
             with console.capture() as capture:
                 rich.inspect(
-                    self.cached_obj.selected_cached_obj.obj, console=console, all=True
+                    self.explorer_layout.selected_object.obj, console=console, all=True
                 )
             str_out = capture.get()
             pydoc.pager(str_out)
@@ -280,21 +296,33 @@ class Explorer:
 
         # Return selected object
         elif key == "r":
-            return self.cached_obj.selected_cached_obj.obj
+            return self.explorer_layout.selected_object.obj
 
     def get_explorer_layout(self) -> Layout:
         if self.stack.visible:
             layout = Layout()
             layout.split_column(
                 self.explorer_layout(
-                    term_width=self.term.width, term_height=self.term.height
+                    term_width=self.term.width,
+                    term_height=self.term.height,
                 ),
                 self.stack(term_width=self.term.width),
             )
             return layout
+        elif self.filter_layout.visible:
+            layout = Layout()
+            layout.split_column(
+                self.explorer_layout(
+                    term_width=self.term.width,
+                    term_height=self.term.height,
+                ),
+                self.filter_layout(),
+            )
+            return layout
         else:
             return self.explorer_layout(
-                term_width=self.term.width, term_height=self.term.height
+                term_width=self.term.width,
+                term_height=self.term.height,
             )
 
     def get_overview_layout(self) -> Layout:
@@ -302,7 +330,7 @@ class Explorer:
             return self.help_layout()
         else:
             return self.overview_layout(
-                cached_obj=self.cached_obj.selected_cached_obj,
+                cached_obj=self.explorer_layout.selected_object,
                 term_height=self.term.height,
                 console=console,
             )
@@ -314,7 +342,11 @@ class Explorer:
 
         layout.split_row(self.get_explorer_layout(), self.get_overview_layout())
 
-        title = self.cached_obj.title
+        title = (
+            self.cached_obj.dotpath
+            + Text(" | ", style="white")
+            + self.cached_obj.typeof
+        )
 
         object_explorer = Panel(
             layout,
@@ -322,7 +354,7 @@ class Explorer:
             subtitle=(
                 "[red][u]q[/u]:quit[/red] "
                 f"[cyan][u]?[/u]:{'exit ' if self.help_layout.visible else ''}help[/] "
-                "[white][dim][u]o[/u]:toggle stack view[/dim]"
+                "[white][dim][u]o[/u]:stack [u]n[/u]:filter [u]r[/u]:return[/dim]"
             ),
             subtitle_align="left",
             height=self.term.height - 1,
@@ -332,7 +364,8 @@ class Explorer:
 
     @property
     def panel_height(self) -> int:
-        # TODO this shouldn't be here
+        # TODO think about putting this somewhere else?
+        # TODO maybe a custom terminal object
         if self.stack.visible:
             return (self.term.height - 10) // 2
         else:
@@ -342,17 +375,20 @@ class Explorer:
 def explore(obj: Any) -> Any:
     """ Run the explorer on the given object """
     try:
-        e = Explorer(obj)
+        frame = inspect.currentframe()
+        name = frame.f_back.f_code.co_names[1]  # type: ignore
+        e = Explorer(obj, name_of_obj=name)
         return e.explore()
-    except Exception as e:
+    except Exception as err:
         console.print_exception(show_locals=True)
         print()
         rich.print(f"[red]{random_error_quote()}")
-        formatted_link = f"https://github.com/kylepollina/objexplore/issues/new?assignees=&labels=&template=bug_report.md&title={e}".replace(
+        formatted_link = f"https://github.com/kylepollina/objexplore/issues/new?assignees=&labels=&template=bug_report.md&title={err}".replace(
             " ", "+"
         )
         print("Please report the issue here:")
         rich.print(f"   [link={formatted_link}][u]{formatted_link}[/u][/link]")
+        print()
         rich.print(
-            "[yellow italic]Make sure to copy/paste the above traceback to the issue to make the issue quicker to solve!"
+            "[yellow italic]Make sure to copy/paste the above traceback to the issue page to make this quicker to fix :)"
         )
